/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.novabpm.identity.auth;

import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.remote.JMXPrincipal;
import javax.security.auth.Subject;

import org.ow2.novabpm.util.Misc;

/**
 * @author "Pierre Vigneras"
 * @date Dec 5, 2007
 */
public abstract class SecurityContext {
  // Singleton!
  private static InheritableThreadLocal<Subject> subjects = new InheritableThreadLocal<Subject>();
  private static final Logger log = Logger.getLogger(SecurityContext.class.getName());

  public static Subject getSubject() {
    if (log.isLoggable(Level.FINEST)) log.entering(SecurityContext.class.getName(), "getSubject");
    final Subject subject = SecurityContext.subjects.get(); 
    Misc.badStateIfNull(subject, "Subject has not been set up using setSubject()!");
    if (log.isLoggable(Level.FINEST)) log.exiting(SecurityContext.class.getName(), "getSubject", subject);
    return subject;
  }

  static void setSubject(final Subject subject) {
    if (log.isLoggable(Level.FINEST)) log.entering(SecurityContext.class.getName(), "setSubject", subject);
	  Misc.checkArgsNotNull(subject, "Subject can't be null!");
	  final Subject current = SecurityContext.subjects.get();
	  if (current != null) {
	 // TODO: This check is required if we consider that one has 
	    if (log.isLoggable(Level.WARNING)) log.warning("Subject is already set to: " + current + ". Did you logout before login?");
    //Misc.badStateIfNotNull(current, "Already setup! Current subject is: " + current + ", passed subject is: " + subject);
	  }
    SecurityContext.subjects.set(subject);
    log.fine("Subject set: " + subject);
    if (log.isLoggable(Level.FINEST)) log.exiting(SecurityContext.class.getName(), "setSubject");
  }

  static void clearSubject() {
    if (log.isLoggable(Level.FINEST)) log.entering(SecurityContext.class.getName(), "clearSubject");
    final Subject subject = SecurityContext.subjects.get();
    if (subject == null) {
      if (log.isLoggable(Level.WARNING)) log.warning("clearSubject() called but subject is already null!");
    }else {
      SecurityContext.subjects.remove();
    }
    log.fine("Subject cleared");
    if (log.isLoggable(Level.FINEST)) log.exiting(SecurityContext.class.getName(), "clearSubject");
  }

  public static String getCallerId() {
    final Subject subject = SecurityContext.getSubject();
    // We use JMXPrincipal as it is the single Principal implementation available 
    // in the whole J2SE 5.
    final Set<JMXPrincipal> principals = subject.getPrincipals(JMXPrincipal.class);
    if (principals == null || principals.size() == 0) {
      throw new RuntimeException("Ouch! Subject does not contain any JMXPrincipal! Subject is: "
              + subject);
    }
    if (principals.size() > 1) {
      throw new RuntimeException("Ouch! Subject contains more than one JMXPrincipal! Subject is: " + subject);
    }
    return principals.toArray(new JMXPrincipal[0])[0].getName();
  }
}
